VS_VERSION_INFO VERSIONINFO
 FILEVERSION 	8,0,0,0
 PRODUCTVERSION 8,0,0,0
 FILEFLAGSMASK 	0x3fL
 FILEFLAGS 	0x0L
 FILEOS 	0x4L
 FILETYPE 	0x2L
 FILESUBTYPE 	0x0L
BEGIN
    BLOCK "StringFileInfo"
    BEGIN
        BLOCK "040904b0"
        BEGIN
            VALUE "FileDescription", "TclX shell\0"
            VALUE "OriginalFilename", "tcl.exe\0"
            VALUE "CompanyName", "TclX Project\0"
            VALUE "FileVersion", 8,0,0,0
            VALUE "LegalCopyright", "Copyright \251 1995-1997\0"
            VALUE "ProductName", "TclX 8.0.0 for Windows\0"
            VALUE "ProductVersion", 8,0,0,0
        END
    END
    BLOCK "VarFileInfo"
    BEGIN
        VALUE "Translation", 0x409, 1200
    END
END




